@@ -30,7 +30,7 @@
 #include <glib.h>
 #include <gdbus.h>
 
-#include "logging.h"
+#include "log.h"
 #include "transfer.h"
 #include "session.h"
 #include "pbap.h"
@@ -246,7 +246,7 @@ static gint pbap_set_path(struct session_data *session, const char *path)
 	if (!path)
 		return OBEX_RSP_BAD_REQUEST;
 
-	if (pbapdata->path != NULL && 	g_str_equal(pbapdata->path, path))
+	if (pbapdata->path != NULL && g_str_equal(pbapdata->path, path))
 		return 0;
 
 	if (gw_obex_chdir(session->obex, "", &err) == FALSE) {
@@ -331,12 +331,15 @@ static void read_return_apparam(struct session_data *session,
 }
 
 static void pull_phonebook_callback(struct session_data *session,
-					void *user_data)
+					GError *err, void *user_data)
 {
 	struct transfer_data *transfer = session->pending->data;
 	DBusMessage *reply;
 	char *buf = "";
 
+	if (session->msg == NULL)
+		goto done;
+
 	reply = dbus_message_new_method_return(session->msg);
 
 	if (transfer->filled > 0)
@@ -350,16 +353,22 @@ static void pull_phonebook_callback(struct session_data *session,
 	g_dbus_send_message(session->conn, reply);
 	dbus_message_unref(session->msg);
 	session->msg = NULL;
+
+done:
+	transfer_unregister(transfer);
 }
 
 static void phonebook_size_callback(struct session_data *session,
-					void *user_data)
+					GError *err, void *user_data)
 {
 	struct transfer_data *transfer = session->pending->data;
 	DBusMessage *reply;
 	guint16 phone_book_size;
 	guint8 new_missed_calls;
 
+	if (session->msg == NULL)
+		goto done;
+
 	reply = dbus_message_new_method_return(session->msg);
 
 	read_return_apparam(session, &phone_book_size, &new_missed_calls);
@@ -372,10 +381,13 @@ static void phonebook_size_callback(struct session_data *session,
 	g_dbus_send_message(session->conn, reply);
 	dbus_message_unref(session->msg);
 	session->msg = NULL;
+
+done:
+	transfer_unregister(transfer);
 }
 
 static void pull_vcard_listing_callback(struct session_data *session,
-					void *user_data)
+					GError *err, void *user_data)
 {
 	struct transfer_data *transfer = session->pending->data;
 	GMarkupParseContext *ctxt;
@@ -383,6 +395,9 @@ static void pull_vcard_listing_callback(struct session_data *session,
 	DBusMessageIter iter, array;
 	int i;
 
+	if (session->msg == NULL)
+		goto complete;
+
 	reply = dbus_message_new_method_return(session->msg);
 
 	if (transfer->filled == 0)
@@ -412,6 +427,8 @@ done:
 	g_dbus_send_message(session->conn, reply);
 	dbus_message_unref(session->msg);
 	session->msg = NULL;
+complete:
+	transfer_unregister(transfer);
 }
 
 static DBusMessage *pull_phonebook(struct session_data *session,
@@ -710,7 +727,8 @@ static DBusMessage *pbap_pull_vcard(DBusConnection *connection,
 
 	if (!pbapdata->path)
 		return g_dbus_create_error(message,
-				ERROR_INF ".Forbidden", "Call Select first of all");
+				ERROR_INF ".Forbidden",
+				"Call Select first of all");
 
 	if (dbus_message_get_args(message, NULL,
 			DBUS_TYPE_STRING, &name,
