@@ -40,11 +40,12 @@
 #include <bluetooth/sdp.h>
 #include <bluetooth/sdp_lib.h>
 
-#include "logging.h"
+#include "log.h"
 #include "pbap.h"
 #include "sync.h"
 #include "transfer.h"
 #include "session.h"
+#include "btio.h"
 
 #define AGENT_INTERFACE  "org.openobex.Agent"
 
@@ -53,6 +54,8 @@
 
 #define FTP_INTERFACE  "org.openobex.FileTransfer"
 
+#define OBEX_IO_ERROR obex_io_error_quark()
+
 static guint64 counter = 0;
 
 static unsigned char pcsuite_uuid[] = { 0x00, 0x00, 0x50, 0x05, 0x00, 0x00,
@@ -71,7 +74,7 @@ struct session_callback {
 	void *data;
 };
 
-struct agent_pending {
+struct pending_data {
 	DBusPendingCall *call;
 	session_callback_t cb;
 	struct transfer_data *transfer;
@@ -81,21 +84,27 @@ struct agent_data {
 	char *name;
 	char *path;
 	guint watch;
-	struct agent_pending *pending;
+	struct pending_data *pending;
 };
 
-static void session_prepare_put(struct session_data *session, void *data);
+static void session_prepare_put(struct session_data *session, GError *err,
+								void *data);
+
+static GQuark obex_io_error_quark(void)
+{
+	return g_quark_from_static_string("obex-io-error-quark");
+}
 
 struct session_data *session_ref(struct session_data *session)
 {
 	g_atomic_int_inc(&session->refcount);
 
-	debug("session_ref(%p): ref=%d", session, session->refcount);
+	DBG("%p: ref=%d", session, session->refcount);
 
 	return session;
 }
 
-static void free_pending(struct agent_pending *pending)
+static void free_pending(struct pending_data *pending)
 {
 	if (pending->call)
 		dbus_pending_call_unref(pending->call);
@@ -156,12 +165,12 @@ static void session_unregistered(struct session_data *session)
 	g_dbus_unregister_interface(session->conn, session->path,
 					SESSION_INTERFACE);
 
-	debug("Session(%p) unregistered %s", session, session->path);
+	DBG("Session(%p) unregistered %s", session, session->path);
 }
 
 static void session_free(struct session_data *session)
 {
-	debug("session_free(%p)", session);
+	DBG("%p", session);
 
 	if (session->agent)
 		agent_release(session);
@@ -172,8 +181,10 @@ static void session_free(struct session_data *session)
 	if (session->obex != NULL)
 		gw_obex_close(session->obex);
 
-	if (session->sock > 2)
-		close(session->sock);
+	if (session->io != NULL) {
+		g_io_channel_shutdown(session->io, TRUE, NULL);
+		g_io_channel_unref(session->io);
+	}
 
 	if (session->path)
 		session_unregistered(session);
@@ -194,7 +205,7 @@ void session_unref(struct session_data *session)
 
 	ret = g_atomic_int_dec_and_test(&session->refcount);
 
-	debug("session_unref(%p): ref=%d", session, session->refcount);
+	DBG("%p: ref=%d", session, session->refcount);
 
 	if (ret == FALSE)
 		return;
@@ -202,88 +213,57 @@ void session_unref(struct session_data *session)
 	session_free(session);
 }
 
-static gboolean rfcomm_callback(GIOChannel *io, GIOCondition cond,
-							gpointer user_data)
+static void rfcomm_callback(GIOChannel *io, GError *err, gpointer user_data)
 {
 	struct callback_data *callback = user_data;
 	struct session_data *session = callback->session;
 	GwObex *obex;
 	int fd;
 
-	if (cond & (G_IO_NVAL | G_IO_ERR))
+	if (err != NULL) {
+		error("%s", err->message);
 		goto done;
+	}
+
+	/* do not close when gw_obex is using the fd */
+	g_io_channel_set_close_on_unref(session->io, FALSE);
+	g_io_channel_unref(session->io);
+	session->io = NULL;
 
 	fd = g_io_channel_unix_get_fd(io);
 
 	obex = gw_obex_setup_fd(fd, session->target,
 			session->target_len, NULL, NULL);
 
-	callback->session->sock = fd;
-	callback->session->obex = obex;
+	session->obex = obex;
 
 done:
-	callback->func(callback->session, callback->data);
+	callback->func(callback->session, err, callback->data);
 
 	session_unref(callback->session);
 
 	g_free(callback);
-
-	return FALSE;
 }
 
-static int rfcomm_connect(const bdaddr_t *src,
-				const bdaddr_t *dst, uint8_t channel,
-					GIOFunc function, gpointer user_data)
+static GIOChannel *rfcomm_connect(const bdaddr_t *src, const bdaddr_t *dst,
+					uint8_t channel, BtIOConnect function,
+					gpointer user_data)
 {
 	GIOChannel *io;
-	struct sockaddr_rc addr;
-	int sk;
+	GError *err = NULL;
 
-	sk = socket(PF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM);
-	if (sk < 0)
-		return -EIO;
+	io = bt_io_connect(BT_IO_RFCOMM, function, user_data, NULL, &err,
+				BT_IO_OPT_SOURCE_BDADDR, src,
+				BT_IO_OPT_DEST_BDADDR, dst,
+				BT_IO_OPT_CHANNEL, channel,
+				BT_IO_OPT_SEC_LEVEL, BT_IO_SEC_LOW,
+				BT_IO_OPT_INVALID);
+	if (io != NULL)
+		return io;
 
-	memset(&addr, 0, sizeof(addr));
-	addr.rc_family = AF_BLUETOOTH;
-	bacpy(&addr.rc_bdaddr, src);
-
-	if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
-		close(sk);
-		return -EIO;
-	}
-
-	io = g_io_channel_unix_new(sk);
-	if (io == NULL) {
-		close(sk);
-		return -ENOMEM;
-	}
-
-	if (g_io_channel_set_flags(io, G_IO_FLAG_NONBLOCK,
-						NULL) != G_IO_STATUS_NORMAL) {
-		g_io_channel_unref(io);
-		close(sk);
-		return -EPERM;
-	}
-
-	memset(&addr, 0, sizeof(addr));
-	addr.rc_family = AF_BLUETOOTH;
-	bacpy(&addr.rc_bdaddr, dst);
-	addr.rc_channel = channel;
-
-	if (connect(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
-		if (errno != EAGAIN && errno != EINPROGRESS) {
-			g_io_channel_unref(io);
-			close(sk);
-			return -EIO;
-		}
-	}
-
-	g_io_add_watch(io, G_IO_OUT | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
-							function, user_data);
-
-	g_io_channel_unref(io);
-
-	return 0;
+	error("%s", err->message);
+	g_error_free(err);
+	return NULL;
 }
 
 static void search_callback(uint8_t type, uint16_t status,
@@ -293,6 +273,7 @@ static void search_callback(uint8_t type, uint16_t status,
 	unsigned int scanned, bytesleft = size;
 	int seqlen = 0;
 	uint8_t dataType, channel = 0;
+	GError *gerr = NULL;
 
 	if (status || type != SDP_SVC_SEARCH_ATTR_RSP)
 		goto failed;
@@ -343,8 +324,11 @@ static void search_callback(uint8_t type, uint16_t status,
 
 	callback->session->channel = channel;
 
-	if (rfcomm_connect(&callback->session->src, &callback->session->dst,
-					channel, rfcomm_callback, callback) == 0) {
+	callback->session->io = rfcomm_connect(&callback->session->src,
+						&callback->session->dst,
+						channel, rfcomm_callback,
+						callback);
+	if (callback->session->io != NULL) {
 		sdp_close(callback->sdp);
 		return;
 	}
@@ -352,7 +336,11 @@ static void search_callback(uint8_t type, uint16_t status,
 failed:
 	sdp_close(callback->sdp);
 
-	callback->func(callback->session, callback->data);
+	g_set_error(&gerr, OBEX_IO_ERROR, -EIO,
+					"Unable to find service record");
+	callback->func(callback->session, gerr, callback->data);
+	g_clear_error(&gerr);
+
 	session_unref(callback->session);
 	g_free(callback);
 }
@@ -377,6 +365,7 @@ static gboolean service_callback(GIOChannel *io, GIOCondition cond,
 	struct callback_data *callback = user_data;
 	sdp_list_t *search, *attrid;
 	uint32_t range = 0x0000ffff;
+	GError *gerr = NULL;
 
 	if (cond & (G_IO_NVAL | G_IO_ERR))
 		goto failed;
@@ -405,7 +394,11 @@ static gboolean service_callback(GIOChannel *io, GIOCondition cond,
 failed:
 	sdp_close(callback->sdp);
 
-	callback->func(callback->session, callback->data);
+	g_set_error(&gerr, OBEX_IO_ERROR, -EIO,
+					"Unable to find service record");
+	callback->func(callback->session, gerr, callback->data);
+	g_clear_error(&gerr);
+
 	session_unref(callback->session);
 	g_free(callback);
 	return FALSE;
@@ -452,7 +445,6 @@ struct session_data *session_create(const char *source,
 		return NULL;
 
 	session->refcount = 1;
-	session->sock = -1;
 	session->channel = channel;
 
 	session->conn = dbus_bus_get(DBUS_BUS_SESSION, NULL);
@@ -499,8 +491,11 @@ struct session_data *session_create(const char *source,
 	callback->data = user_data;
 
 	if (session->channel > 0) {
-		err = rfcomm_connect(&session->src, &session->dst,
-				session->channel, rfcomm_callback, callback);
+		session->io = rfcomm_connect(&session->src, &session->dst,
+							session->channel,
+							rfcomm_callback,
+							callback);
+		err = (session->io == NULL) ? -EINVAL : 0;
 	} else {
 		callback->sdp = service_connect(&session->src, &session->dst,
 						service_callback, callback);
@@ -520,7 +515,7 @@ void session_shutdown(struct session_data *session)
 {
 	struct transfer_data *transfer;
 
-	debug("session_shutdown(%p)", session);
+	DBG("%p", session);
 	transfer = session->pending ? session->pending->data : NULL;
 
 	session_ref(session);
@@ -751,7 +746,7 @@ static const GMarkupParser parser = {
 };
 
 static void list_folder_callback(struct session_data *session,
-					void *user_data)
+					GError *err, void *user_data)
 {
 	struct transfer_data *transfer = session->pending->data;
 	GMarkupParseContext *ctxt;
@@ -791,7 +786,8 @@ done:
 	session->msg = NULL;
 }
 
-static void get_file_callback(struct session_data *session, void *user_data)
+static void get_file_callback(struct session_data *session, GError *err,
+							void *user_data)
 {
 
 }
@@ -800,7 +796,7 @@ static void session_request_reply(DBusPendingCall *call, gpointer user_data)
 {
 	struct session_data *session = user_data;
 	struct agent_data *agent = session->agent;
-	struct agent_pending *pending = agent->pending;
+	struct pending_data *pending = agent->pending;
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	const char *name;
 	DBusError derr;
@@ -810,29 +806,39 @@ static void session_request_reply(DBusPendingCall *call, gpointer user_data)
 		error("Replied with an error: %s, %s",
 				derr.name, derr.message);
 		dbus_error_free(&derr);
-		goto fail;
+		dbus_message_unref(reply);
+		transfer_unregister(pending->transfer);
+		return;
 	}
 
 	dbus_message_get_args(reply, NULL,
 			DBUS_TYPE_STRING, &name,
 			DBUS_TYPE_INVALID);
 
-	debug("Agent.Request() reply: %s", name);
+	DBG("Agent.Request() reply: %s", name);
 
 	if (strlen(name)) {
 		g_free(pending->transfer->name);
 		pending->transfer->name = g_strdup(name);
 	}
 
-	pending->cb(session, pending->transfer);
-
+	pending->cb(session, NULL, pending->transfer);
+	dbus_message_unref(reply);
 	free_pending(pending);
 	agent->pending = NULL;
 
 	return;
+}
 
-fail:
-	transfer_unregister(pending->transfer);
+static gboolean session_request_proceed(gpointer data)
+{
+	struct pending_data *pending = data;
+	struct transfer_data *transfer = pending->transfer;
+
+	pending->cb(transfer->session, NULL, transfer);
+	free_pending(pending);
+
+	return FALSE;
 }
 
 static int session_request(struct session_data *session, session_callback_t cb,
@@ -840,11 +846,14 @@ static int session_request(struct session_data *session, session_callback_t cb,
 {
 	struct agent_data *agent = session->agent;
 	DBusMessage *message;
-	DBusPendingCall *call;
-	struct agent_pending *pending;
+	struct pending_data *pending;
+
+	pending = g_new0(struct pending_data, 1);
+	pending->cb = cb;
+	pending->transfer = transfer;
 
 	if (agent == NULL || transfer->path == NULL) {
-		cb(session, transfer);
+		g_idle_add(session_request_proceed, pending);
 		return 0;
 	}
 
@@ -855,34 +864,32 @@ static int session_request(struct session_data *session, session_callback_t cb,
 			DBUS_TYPE_OBJECT_PATH, &transfer->path,
 			DBUS_TYPE_INVALID);
 
-
-	if (!dbus_connection_send_with_reply(session->conn, message, &call, -1)) {
+	if (!dbus_connection_send_with_reply(session->conn, message,
+						&pending->call, -1)) {
 		dbus_message_unref(message);
 		return -ENOMEM;
 	}
 
-	dbus_message_unref(message);
-
-	pending = g_new0(struct agent_pending, 1);
-	pending->call = call;
-	pending->cb = cb;
-	pending->transfer = transfer;
 	agent->pending = pending;
 
-	dbus_pending_call_set_notify(call, session_request_reply, session, NULL);
+	dbus_message_unref(message);
 
-	debug("Agent.Request(\"%s\")", transfer->path);
+	dbus_pending_call_set_notify(pending->call, session_request_reply,
+					session, NULL);
+
+	DBG("Agent.Request(\"%s\")", transfer->path);
 
 	return 0;
 }
 
 static void session_terminate_transfer(struct session_data *session,
-					struct transfer_data *transfer)
+					struct transfer_data *transfer,
+					GError *gerr)
 {
 	struct session_callback *callback = session->callback;
 
 	if (callback) {
-		callback->func(session, callback->data);
+		callback->func(session, gerr, callback->data);
 		return;
 	}
 
@@ -921,14 +928,14 @@ static void session_notify_complete(struct session_data *session,
 
 done:
 
-	debug("Transfer(%p) complete", transfer);
+	DBG("Transfer(%p) complete", transfer);
 
-	session_terminate_transfer(session, transfer);
+	session_terminate_transfer(session, transfer, NULL);
 }
 
 static void session_notify_error(struct session_data *session,
 				struct transfer_data *transfer,
-				const char *err)
+				GError *err)
 {
 	struct agent_data *agent = session->agent;
 	DBusMessage *message;
@@ -938,7 +945,7 @@ static void session_notify_error(struct session_data *session,
 
 		reply = g_dbus_create_error(session->msg,
 					"org.openobex.Error.Failed",
-					err);
+					err->message);
 		g_dbus_send_message(session->conn, reply);
 
 		dbus_message_unref(session->msg);
@@ -957,15 +964,15 @@ static void session_notify_error(struct session_data *session,
 
 	dbus_message_append_args(message,
 			DBUS_TYPE_OBJECT_PATH, &transfer->path,
-			DBUS_TYPE_STRING, &err,
+			DBUS_TYPE_STRING, &err->message,
 			DBUS_TYPE_INVALID);
 
 	g_dbus_send_message(session->conn, message);
 
 done:
-	error("Transfer(%p) Error: %s", transfer, err);
+	error("Transfer(%p) Error: %s", transfer, err->message);
 
-	session_terminate_transfer(session, transfer);
+	session_terminate_transfer(session, transfer, err);
 }
 
 static void session_notify_progress(struct session_data *session,
@@ -1004,7 +1011,7 @@ static void session_notify_progress(struct session_data *session,
 	g_dbus_send_message(session->conn, message);
 
 done:
-	debug("Transfer(%p) progress: %ld bytes", transfer,
+	DBG("Transfer(%p) progress: %ld bytes", transfer,
 			(long int ) transferred);
 
 	if (transferred == transfer->size)
@@ -1015,6 +1022,7 @@ static void transfer_progress(struct transfer_data *transfer, gint64 transferred
 				int err, void *user_data)
 {
 	struct session_data *session = user_data;
+	GError *gerr = NULL;
 
 	if (err != 0)
 		goto fail;
@@ -1024,38 +1032,57 @@ static void transfer_progress(struct transfer_data *transfer, gint64 transferred
 	return;
 
 fail:
-	session_notify_error(session, transfer,
+	g_set_error(&gerr, OBEX_IO_ERROR, err, "%s",
 			err > 0 ? OBEX_ResponseToString(err) : strerror(-err));
+	session_notify_error(session, transfer, gerr);
+	g_clear_error(&gerr);
 }
 
-static void session_prepare_get(struct session_data *session, void *data)
+static void session_prepare_get(struct session_data *session,
+				GError *err, void *data)
 {
 	struct transfer_data *transfer = data;
+	int ret;
 
-	if (transfer_get(transfer, transfer_progress, session) < 0)
-		transfer_unregister(transfer);
+	ret = transfer_get(transfer, transfer_progress, session);
+	if (ret < 0) {
+		GError *gerr = NULL;
+
+		g_set_error(&gerr, OBEX_IO_ERROR, ret, "%s", strerror(-ret));
+		session_notify_error(session, transfer, gerr);
+		g_clear_error(&gerr);
+		return;
+	}
+
+	DBG("Transfer(%p) started", transfer);
 }
 
 int session_get(struct session_data *session, const char *type,
 		const char *filename, const char *targetname,
-		const guint8  *apparam, gint apparam_size,
+		const guint8 *apparam, gint apparam_size,
 		session_callback_t func)
 {
 	struct transfer_data *transfer;
-	struct transfer_params *params;
+	struct transfer_params *params = NULL;
 	int err;
 
 	if (session->obex == NULL)
 		return -ENOTCONN;
 
-	params = g_new0(struct transfer_params, 1);
-	params->data = apparam;
-	params->size = apparam_size;
+	if (apparam != NULL) {
+		params = g_new0(struct transfer_params, 1);
+		params->data = g_new(guint8, apparam_size);
+		memcpy(params->data, apparam, apparam_size);
+		params->size = apparam_size;
+	}
 
 	transfer = transfer_register(session, filename, targetname, type,
 					params);
 	if (transfer == NULL) {
-		g_free(params);
+		if (params != NULL) {
+			g_free(params->data);
+			g_free(params);
+		}
 		return -EIO;
 	}
 
@@ -1332,7 +1359,7 @@ int session_register(struct session_data *session)
 		return -EIO;
 	}
 
-	debug("Session(%p) registered %s", session, session->path);
+	DBG("Session(%p) registered %s", session, session->path);
 
 	return 0;
 }
@@ -1347,12 +1374,24 @@ void session_set_data(struct session_data *session, void *priv)
 	session->priv = priv;
 }
 
-static void session_prepare_put(struct session_data *session, void *data)
+static void session_prepare_put(struct session_data *session,
+				GError *err, void *data)
 {
 	struct transfer_data *transfer = data;
+	int ret;
 
-	if (transfer_put(transfer, transfer_progress, session) < 0)
-		transfer_unregister(transfer);
+	ret = transfer_put(transfer, transfer_progress, session);
+	if (ret < 0) {
+		GError *gerr = NULL;
+
+		g_set_error(&gerr, OBEX_IO_ERROR, ret, "%s (%d)",
+							strerror(-ret), -ret);
+		session_notify_error(session, transfer, gerr);
+		g_clear_error(&gerr);
+		return;
+	}
+
+	DBG("Transfer(%p) started", transfer);
 }
 
 int session_put(struct session_data *session, char *buf, const char *targetname)
